/*
    Copyright (C) 1995-2001 AROS - The Amiga Research OS
    $Id: modlist.c,v 1.5 2005/12/02 13:42:43 laire Exp $
 
    Desc:
    Lang: english
*/

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <exec/tasks.h>
#include <exec/execbase.h>
#include <exec/resident.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <aros/machine.h>
#include <string.h>

#include "modlist_version.h"

static const char version[] = VERSTAG;

#define ARG_TEMPLATE "VERBOSE/S"

enum
{
	ARG_VERBOSE = 0,
	NOOFARGS
};

struct res
{
	struct Resident	*address;
	struct Resident	res;
};


static int fillbuffer(struct res **buffer, IPTR size, struct ExecBase *SysBase);


static const
char	*Flags[]=
{
	"ColdStart",
	"SingleTask",
	"AfterDos",
	"PPC",
	"Bit4",
	"*Asynchron*",
	"Extended",
	"AutoInit"
};

int main(void)
{
	struct ExecBase		*SysBase;
	struct DosLibrary	*DOSBase;
	IPTR           args[NOOFARGS] = { (IPTR)NULL};
	struct RDArgs *rda;
	IPTR size;
	struct res *buffer,*ress,*ress2;
	int	i;
	char	buf[256];
	LONG error=RETURN_OK;

	SysBase=*((struct ExecBase**) 4);
	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37))!=0)
	{
		rda=(struct RDArgs *)ReadArgs(ARG_TEMPLATE, args, NULL);

		if (rda != NULL)
		{
			for(size=4096;;size+=2048)
			{
				buffer=AllocVec(size,MEMF_ANY);
				if(buffer==NULL)
				{
					FPuts(Output(),"Not Enough memory for resident buffer\n");
					CloseLibrary((struct Library *) DOSBase);
					return RETURN_FAIL;
				}
				ress=buffer;
				if(fillbuffer(&ress,size,SysBase))
				{
					FPuts(Output(),"address                    name                version   pri    flags\n"
					      "--------------------------------------------------------------------------\n");
					for(ress2=buffer;ress2<ress;ress2++)
					{
						LONG revision = -1;
						LONG hasrev;

						if (ress2->res.rt_Name)
						{
							strncpy(buf,ress2->res.rt_Name,sizeof(buf)-1);
							buf[sizeof(buf)-1]	=	'\0';
							for (i=0;i<sizeof(buf);i++)
							{
								if (buf[i]=='\0')
								{
									break;
								}
								else
								if (buf[i]=='\n')
								{
									buf[i]	=	'\0';
									break;
								}
								else
								if (buf[i]=='\r')
								{
									buf[i]	=	'\0';
									break;
								}
							}
						}
						else
						{
							buf[0]	=	'\0';
						}

						if(ress2->res.rt_Flags & RTF_EXTENDED)
						{
							revision = ress2->res.rt_Revision;
							hasrev = TRUE;
						}
						else
						{
							/* Scan rt_IdString for revision - Piru */
							CONST_STRPTR ids = ress2->res.rt_IdString;
							hasrev = FALSE;
							if (ids)
							{
								LONG tmp;
								LONG res;

								while (*ids && *ids != ' ')
								{
									ids++;
								}
								res = StrToLong(ids, &tmp);
								ids += res;
								if (res > 0 && *ids++ == '.')
								{
									if (StrToLong(ids, &tmp) > 0)
									{
										revision = tmp;
										hasrev = TRUE;
									}
								}
							}
						}

						if (hasrev)
						{
							Printf("0x%08.lx\t%30.s\t%3lu.%-3lu\t%4.ld\t<",
								(ULONG)ress2->address,
								(ULONG)buf,
								ress2->res.rt_Version,
								revision,
								ress2->res.rt_Pri);
						}
						else
						{
							Printf("0x%08.lx\t%30.s\t%3lu\t%4.ld\t<",
								(ULONG)ress2->address,
								(ULONG)buf,
								ress2->res.rt_Version,
								ress2->res.rt_Pri);
						}


						if (ress2->res.rt_Flags)
						{
							int flag=FALSE;
							for (i=7;i>=0;i--)
							{
								if (ress2->res.rt_Flags & 1UL<<i)
								{
									if (flag)
									{
										Printf(" | ");
									}
									Printf(Flags[i]);
									flag = TRUE;
								}
							}
						}
						else
						{
							Printf("NEVER");
						}
						Printf(">\n");

						if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
						{
							error = RETURN_FAIL;
							SetIoErr(ERROR_BREAK);
							break;
						}
					}
					FreeVec(buffer);
					break;
				}
				FreeVec(buffer);
			}
			FreeArgs(rda);
		}
		else
		{
			error = RETURN_FAIL;
		}

		if (error != RETURN_OK)
		{
			PrintFault(IoErr(), NULL);
		}
		CloseLibrary((struct Library *)DOSBase);
	}

	return error;
}

static int addres(struct Resident *r, struct res **l, STRPTR *e)
{
	if((STRPTR)(*l+1) > *e)
		return 0;

	(*l)->address = r;

	memcpy(&((*l)->res),
               r,
               sizeof(*r));

	++*l;
	return 1;
}

static int fillbuffer(struct res **buffer, IPTR size, struct ExecBase *SysBase)
{
	struct Resident *MyResident;
	struct Resident **ResPtr;
	STRPTR end=(STRPTR)*buffer+size;
	int ret=1;
	Forbid();


	ResPtr=(struct Resident**) SysBase->ResModules;
	while ((MyResident=*ResPtr++))
	{
		/* Test for bit 31=Signed */
		if (((LONG) MyResident) > 0)
		{
			ret=addres(MyResident,buffer,&end);
		}
		else
		{
			ResPtr	=(struct Resident**) (((ULONG) MyResident) & ~(1<<31));
		}
	}

	Permit();
	return ret;
}

