;/*
sc resopt link optimize debug=line idir=include: nostartup nostackcheck opt data=far Version.c
quit
:ts=4
*/

/*
	(C) 1995-98 AROS - The Amiga Research OS
	(C) 2002-2012 Harry Sintonen
	$Id: version.c,v 1.12 2014/03/31 16:46:25 piru Exp $

	Desc: Version CLI command
	Lang: english
*/

/*
  TODO:
   o pattern for FILE/M ?
   o scanning: if directory is given, scan all entries in it, or recursive if
     ALL is passed (VERSION/N,REVISION/N are ignored here). Problem: how to get
     version of filesystem then?
*/

#define AROS_ALMOST_COMPATIBLE
#include <exec/types.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <exec/resident.h>
#include <exec/libraries.h>
#include <exec/execbase.h>
#include <dos/dos.h>
#include <dos/datetime.h>
#include <dos/dosextens.h>
#include <rexx/storage.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/rexxsyslib.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "md5.h"

#ifdef __SASC
typedef const UBYTE *CONST_STRPTR;
typedef LONG IPTR;
#define ForeachNode(l,n)  \
for (  \
n = (void *)(((struct List *)(l))->lh_Head);  \
((struct Node *)(n))->ln_Succ;  \
n = (void *)(((struct Node *)(n))->ln_Succ)  \
)
#endif /* __SASC */

#ifndef FORMAT_DEF
#define FORMAT_DEF 4
#endif

#include "version_version.h"

static const char version[] = VERSTAG;

//static const UBYTE ERROR_HEADER[] = "Version";
#define ERROR_HEADER NULL
struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;
struct Library		*UtilityBase;

#define TEMPLATE "NAME/M,MD5SUM/S,VERSION/N,REVISION/N,FILE/S,FULL/S,RES/S"
struct
{
	CONST_STRPTR *arg_name;
	IPTR    arg_md5sum;
	LONG   *arg_version;
	LONG   *arg_revision;
	IPTR    arg_file;
	IPTR    arg_full;
	IPTR    arg_res;
}
args = { NULL, NULL, NULL, 0L, 0L, 0L };

LONG mversion, mrevision;

struct
{
	STRPTR pv_name;
	ULONG  pv_flags;
	LONG   pv_version;
	LONG   pv_revision;
	//LONG   pv_days;
	STRPTR pv_vername;
	STRPTR pv_revname;
	STRPTR pv_datestr;
	STRPTR pv_extralf;
	STRPTR pv_extrastr;
	UBYTE  pv_md5sum[16];
}
parsedver = { NULL, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, {0}};

#define PVF_MD5SUM		(1 << 0)
#define PVF_NOVERSION	(1 << 1)

static
int makeverstring(CONST_STRPTR name);
static
void printverstring(void);
static
void freeverstring(void);
static
int makesysver(void);
static
int cmpargsparsed(void);

/******************************* main program ****************************/

int main(void)
{
	LONG error = RETURN_FAIL;

	SysBase	= *((struct ExecBase**) 4);
	DOSBase	= (struct DosLibrary *) OpenLibrary("dos.library", 37);
	UtilityBase	= OpenLibrary("utility.library", 37);
	if (DOSBase && UtilityBase)
	{
		struct RDArgs *rda;

		rda = ReadArgs(TEMPLATE, (IPTR *) &args, NULL);
		if (rda)
		{
			if (!args.arg_name || !*args.arg_name)
			{
				/* No args, make system version */
				error = makesysver();
				if (error == RETURN_OK)
				{
					printverstring();

					if (parsedver.pv_flags & PVF_NOVERSION)
					{
						error = RETURN_FAIL;
					}
					if (error == RETURN_OK)
					{
						error = cmpargsparsed();
					}
				}
				freeverstring();
			}
			else
			{
				CONST_STRPTR *name;
				BOOL multifile;

#if 1
				/* Workaround for:
				 * version file ver
				 * version file ver rev
				 */
				if (!args.arg_version && !args.arg_revision)
				{
					LONG narg = 1;
					while (args.arg_name[narg]) { narg++; }
					if (narg == 2 || narg == 3)
					{
						if (StrToLong(args.arg_name[1], &mversion) > 0)
						{
							args.arg_version = &mversion;

							args.arg_name[1] = args.arg_name[2];
							if (narg == 3)
							{
								args.arg_name[2] = NULL;
							}

							if (narg == 3)
							{
								if (StrToLong(args.arg_name[1], &mrevision) > 0)
								{
									args.arg_revision = &mrevision;
									args.arg_name[1] = NULL;
								}
							}
						}
					}
				}
#endif
				multifile = args.arg_name[1] != NULL;

				for (name = args.arg_name; *name; name++)
				{
					error = makeverstring(*name);
					if (error == RETURN_OK)
					{
#if 0
						if (multifile)
						{
							/* Multiple args, don't print extra info */
							FreeVec(parsedver.pv_extrastr);
							parsedver.pv_extrastr = NULL;
						}
#endif

						printverstring();

						if (!multifile)
						{
							/* Single args, do compare stuff also */
							if (parsedver.pv_flags & PVF_NOVERSION)
							{
								error = RETURN_FAIL;
							}
							if (error == RETURN_OK)
							{
								error = cmpargsparsed();
							}
						}
					}

					freeverstring();
				}
			}

			FreeArgs(rda);
		}
		else
		{
			PrintFault(IoErr(), (STRPTR) ERROR_HEADER);
			error = RETURN_FAIL;
		}
	}

	if (UtilityBase)
	{
		CloseLibrary(UtilityBase);
	}
	if (DOSBase)
	{
		CloseLibrary((struct Library *) DOSBase);
	}

	return error;
}


/**************************** support functions ************************/

/* Duplicate string, by given length or -1 for full length
 */
static
STRPTR dupstr(CONST_STRPTR buffer, LONG len)
{
	STRPTR ret = NULL;

	if (buffer)
	{
		if (len == -1)
			len = strlen(buffer);

		ret = AllocVec(len + 1, MEMF_ANY);
		if (ret)
		{
			CopyMem((STRPTR) buffer, ret, len);
			ret[len] = '\0';
		}
	}

	return ret;
}


static
inline int myisspace(int c)
{
	return (c == ' ' || c == '\t');
}


/* Return a pointer to a string, stripped by all leading whitespace characters
 * (SPACE, TAB).
 */
static
STRPTR skipwhites(CONST_STRPTR buffer)
{
	for (;; buffer++)
	{
		if (buffer[0] == '\0' || !isspace(buffer[0]))
		{
			return (STRPTR) buffer;
		}
	}
}


/* Return a pointer to a string, stripped by all leading space characters
 * (SPACE).
 */
static
STRPTR skipspaces(CONST_STRPTR buffer)
{
	for (;; buffer++)
	{
		if (buffer[0] == '\0' || buffer[0] != ' ')
		{
			return (STRPTR) buffer;
		}
	}
}


/* Strip all whitespace-characters from the end of a string. Note that the
 * buffer passed in will be modified!
 */
static
void stripwhites(STRPTR buffer)
{
	int len = strlen(buffer);

	while (len > 0)
	{
		if (!isspace(buffer[len-1]))
		{
			buffer[len] = '\0';
			return;
		}
		len--;
	}
	buffer[len] = '\0';
}


/* Searches for a given string in a file and stores up to *lenptr characters
 * into the buffer beginning with the first character after the given string.
 */
static
int findinfile(BPTR file, CONST_STRPTR string, STRPTR buffer, int *lenptr, unsigned char digest[16])
{
	int error = RETURN_OK;
	int buflen = *lenptr, len = 0, actual, pos, stringlen;
	BOOL ready = FALSE;
	MD5_CTX md5ctx;
	STRPTR tmp;
	BOOL seekendhack = !args.arg_md5sum;

	tmp = AllocMem(buflen, MEMF_PUBLIC);
	if (!tmp)
	{
		return RETURN_FAIL;
	}

	stringlen = strlen(string);
	*lenptr = -1;

	if (args.arg_md5sum)
	{
		MD5Init(&md5ctx);
	}

	for (;;)
	{
		if (SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
		{
			error = RETURN_FAIL;
			SetIoErr(ERROR_BREAK);
			break;
		}

		actual = Read(file, &tmp[len], buflen - len);
		if (actual == -1 || actual == 0)
		{
			if (actual == -1)
			{
				error = RETURN_FAIL;
			}
			break;
		}

		if (args.arg_md5sum)
		{
			MD5Update(&md5ctx, &tmp[len], actual);
		}

		len += actual;
		pos = 0;

		if (ready)
		{
			/* If we get here we're scanning the rest of the file for md5sum. - Piru */
			len = 0;
		}
		else
		{
			if (seekendhack)
			{
				seekendhack = FALSE;

				/* If this is a version 0 MorphOS binary perform an optimization
				 * trick where we only look at the file end for the version
				 * tag... This should considerably speed up version command on
				 * large executables. - Piru
				 */
				if (len == buflen && memcmp(tmp, "\x7fMOS\0\0", 6) == 0)
				{
					if (Seek64(file, -buflen, OFFSET_END) == -1LL)
					{
						error = RETURN_FAIL;
						goto outta;
					}

					len = 0;
					continue;
				}
			}

			while ((len - pos) >= stringlen)
			{
				/* Compare the current buffer position with the supplied string. */
				if (strncmp(&tmp[pos], string, stringlen) == 0)
				{
					/* It is equal! Now move the rest of the buffer to the top of
					 * the buffer and fill it up.
					 */
					int findstrlen = len - pos - stringlen;

					memcpy(buffer, &tmp[pos + stringlen], findstrlen);

					if (SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
					{
						error = RETURN_FAIL;
						SetIoErr(ERROR_BREAK);
						goto outta;
					}

					len = Read(file, &buffer[findstrlen], buflen - findstrlen);
					if (len >= 0)
					{
						if (args.arg_md5sum)
						{
							MD5Update(&md5ctx, &buffer[findstrlen], len);
						}

						*lenptr = findstrlen + len;
					}
					else
					{
						error = RETURN_FAIL;
					}
					ready = TRUE;
					break;
				}
				pos++;
			}
			/* Move the rest of the buffer that could not be compared (because it
			 * is smaller than the string to compare) to the top of the buffer.
			 */
			if (!ready)
			{
				memmove(tmp, &tmp[len - stringlen], stringlen);
			}
			else
			{
				/* If we're not md5summing, stop file scanning now. - Piru */
				if (!args.arg_md5sum)
				{
					break;
				}
			}
			len = stringlen;
		}
	}
	outta:

	FreeMem(tmp, buflen);

	if (len == -1)
	{
		error = RETURN_FAIL;
	}

	if (args.arg_md5sum)
	{
		memset(digest, 0, 16);
		MD5Final(&md5ctx, digest);
	}

	return error;
}


/*************************** parsing functions *************************/

/* Convert a date in the form DD.MM.YY or DD.MM.YYYY into a numerical
 * value. Return FALSE, if buffer doesn't contain a valid date.
 */
static
BOOL makedatefromstring(CONST_STRPTR *bufptr)
{
	CONST_STRPTR buffer = *bufptr;
	struct DateTime dt;
	CONST_STRPTR headerstart, end;
	STRPTR newbuf;
	LONG res;
	int len, i;
	UBYTE c;

	//if (isspace(buffer[0]))
	//  buffer++;

	headerstart = buffer;

	buffer = strchr(buffer, '(');
	if (!buffer)
	{
		return FALSE;
	}
	buffer++;
	end = strchr(buffer, ')');
	if (!end)
	{
		return FALSE;
	}
	len = (int)(end - buffer);
	newbuf = dupstr(buffer, len);
	if (!newbuf)
	{
		return FALSE;
	}
	for (i = 0; i < len; i++)
	{
		c = newbuf[i];

		if (c == '.' || c == '/')
			newbuf[i] = '-';
		else if (!isalnum(c))
		{
			end = buffer + i;
			newbuf[i] = '\0';
			break;
		}
	}

	//Printf("date: \"%s\"\n", (LONG) newbuf);

	dt.dat_Format  = FORMAT_CDN;
	dt.dat_Flags   = 0;
	dt.dat_StrDay  = NULL;
	dt.dat_StrDate = newbuf;
	dt.dat_StrTime = NULL;
	res = StrToDate(&dt);
	if (!res)
	{
		dt.dat_Format  = FORMAT_DOS;
		res = StrToDate(&dt);
		if (!res)
		{
			//Printf("StrToDate failed!\n");
			FreeVec(newbuf);
			return FALSE;
		}
	}
	FreeVec(newbuf);

	//parsedver.pv_days = dt.dat_Stamp.ds_Days;

	parsedver.pv_datestr = AllocVec(buffer - headerstart + LEN_DATSTRING + 2, MEMF_ANY);
	if (!parsedver.pv_datestr)
	{
		return FALSE;
	}

	dt.dat_Stamp.ds_Minute = 0;
	dt.dat_Stamp.ds_Tick   = 0;

	dt.dat_StrDate = parsedver.pv_datestr + (buffer - headerstart);
	dt.dat_Format  = FORMAT_DEF;
	if (!DateToStr(&dt))
	{
		//Printf("DateToStr failed!\n");
		return FALSE;
	}

	CopyMem((STRPTR) headerstart, parsedver.pv_datestr, buffer - headerstart);
	res = strlen(parsedver.pv_datestr);
	parsedver.pv_datestr[res++] = ')';
	parsedver.pv_datestr[res] = '\0';

	*bufptr = end + 1;

	return TRUE;
}


/* Check whether the given string contains a version in the form
 * <version>.<revision> . If not return FALSE, otherwise fill in parsedver and
 * return TRUE.
 */
static
BOOL makeversionfromstring(CONST_STRPTR *bufptr)
{
	LONG pos, ver, rev;
	CONST_STRPTR buffer = *bufptr;
	CONST_STRPTR verstart, revstart;

	//Printf("makeversionfromstring: buffer \"%s\"\n", (LONG) buffer);

	/* Do version */

	verstart = buffer;
	pos = StrToLong((STRPTR) buffer, &ver);
	if (pos == -1)
	{
		return FALSE;
	}
	parsedver.pv_version  = ver;
	parsedver.pv_revision = -1;

	parsedver.pv_vername = dupstr(verstart, pos);
	if (!parsedver.pv_vername)
	{
		return FALSE;
	}

	/* Do revision */

	buffer += pos;
	revstart = buffer;
	buffer = skipspaces(buffer);  /* NOTE: skipspaces, not skipwhites! */
	if (*buffer != '.')
	{
		*bufptr = buffer;
		return TRUE;
	}

	buffer++;
	pos = StrToLong((STRPTR) buffer, &rev);
	if (pos == -1)
	{
		*bufptr = buffer;
		return TRUE;
	}

	parsedver.pv_revision = rev;

	/* calc the revision string len */
	pos = buffer + pos - revstart;
	parsedver.pv_revname = dupstr(revstart, pos);
	if (!parsedver.pv_revname)
	{
		return FALSE;
	}

	*bufptr = revstart + pos;

	return TRUE;
}


static
void printverstring(void)
{
	if (args.arg_md5sum)
	{
		if (parsedver.pv_flags & PVF_MD5SUM)
		{
			VPrintf("%08lX%08lX%08lX%08lX  ", parsedver.pv_md5sum);
		}
		else
		{
			/*     "01234567012345670123456701234567: " */
			PutStr("<no md5sum available>             ");
		}
	}

	if (parsedver.pv_flags & PVF_NOVERSION)
	{
		Printf("%s\n", (LONG) parsedver.pv_name);
	}
	else
	{
		if (args.arg_full)
		{
			/* If md5sum output was there, avoid linefeed to allow parsing the output - Piru */
			if (args.arg_md5sum)
			{
				parsedver.pv_extralf = " ";
			}

			Printf("%s%s%s%s%s%s%s\n",
			       (LONG) parsedver.pv_name, (LONG) (*parsedver.pv_name ? " " : ""),
			       (LONG) parsedver.pv_vername, (LONG) parsedver.pv_revname,
			       (LONG) parsedver.pv_datestr, (LONG) parsedver.pv_extralf, (LONG) parsedver.pv_extrastr);
		}
		else
		{
			Printf("%s%s%s%s\n",
			       (LONG) parsedver.pv_name, (LONG) (*parsedver.pv_name ? " " : ""),
			       (LONG) parsedver.pv_vername, (LONG) parsedver.pv_revname);
		}
	}
}


static
int makedata(CONST_STRPTR buffer, CONST_STRPTR ptr, int pos)
{
	//Printf("makedata: buffer \"%s\" prt \"%s\"\n", (LONG) buffer, (LONG) ptr);

	if (makeversionfromstring(&ptr))
	{
		CONST_STRPTR endp;
		BOOL doskip;

		/* It is! */
		/* Copy the program-name into a buffer. */
		parsedver.pv_name = dupstr(buffer, pos);
		if (!parsedver.pv_name)
		{
			PrintFault(ERROR_NO_FREE_STORE, (STRPTR) ERROR_HEADER);
			return RETURN_FAIL;
		}

		/* Now find the date */
		//Printf("makedatafromstring: ptr #1: \"%s\"\n", (LONG) ptr);
		doskip = strchr(ptr, '(') ? TRUE : FALSE;
		(void) makedatefromstring(&ptr);

		//Printf("makedatafromstring: ptr #2: \"%s\"\n", (LONG) ptr);
		if (doskip)
			ptr = skipspaces(ptr);	/* NOTE: not skipwhites! */
		for (endp = ptr; *endp != '\0' && *endp != '\r' && *endp != '\n'; endp++)
			;
		pos = endp - ptr;
		if (pos)
		{
			parsedver.pv_extrastr = dupstr(ptr, pos);
			if (!parsedver.pv_extrastr)
			{
				PrintFault(ERROR_NO_FREE_STORE, (STRPTR) ERROR_HEADER);
				return RETURN_FAIL;
			}

			if (doskip)
				parsedver.pv_extralf = "\n";
		}

		return 1;
	}

	return 0;
}


/* Retrieves version information from string. The data is stored in the
 * global struct parsedver.pv_
 */

static
int makedatafromstring(CONST_STRPTR buffer)
{
	int error = RETURN_OK;
	int pos = 0;
	LONG add, dummy;

	//Printf("makedatafromstring: buffer: \"%s\"\n", (LONG) buffer);

#if 0
	if ((add = StrToLong((STRPTR) buffer, &dummy)) != -1)
	{
		if (!makedata(buffer, buffer, 0))
			error = -1;

		return error;
	}
#endif

	while (buffer[pos] && buffer[pos] != '\r' && buffer[pos] != '\n')
	{
		/* NOTE: Not isspace()! - Piru */
		if (myisspace(buffer[pos]) &&
		    (add = StrToLong((STRPTR) buffer + pos + 1, &dummy)) != -1)
		{
			CONST_STRPTR ptr;

			/* Found something, which looks like a version. Now check, if it
			 * really is.
			 */

			//Printf("makedatafromstring: buffer + %ld: \"%s\"\n", pos, (LONG) buffer + pos);

			ptr = buffer + pos + 1;

			if (makedata(buffer, ptr, pos))
			{
				break;
			}
		}
		pos++;
	}

	if (!buffer[pos] || buffer[pos] == '\r' || buffer[pos] == '\n')
	{
		CONST_STRPTR endp;

		/* use the whatever is after ver tag as name */
		for (endp = buffer; *endp != '\0' && *endp != '\r' && *endp != '\n'; endp++)
			;
		pos = endp - buffer;

		parsedver.pv_name = dupstr(buffer, pos);
		if (!parsedver.pv_name)
		{
			PrintFault(ERROR_NO_FREE_STORE, (STRPTR) ERROR_HEADER);
			return RETURN_FAIL;
		}
	}

	/* Strip any whitespaces from the tail of the program-name.
	 */
	if (parsedver.pv_name)
	{
		stripwhites(parsedver.pv_name);
	}

	return error;
}


/* Case-insensitive FindResident()
 */
static
struct Resident *findresident(CONST_STRPTR name)
{
	struct Resident **rp;
	struct Resident *resident;

	rp = (struct Resident **) SysBase->ResModules;

	while ((resident = *rp++))
	{
		if (((LONG) resident) > 0)
		{
			if (!Stricmp(resident->rt_Name, (STRPTR) name))
			{
				break;
			}
		}
		else
		{
			rp = (struct Resident **) (((ULONG) resident) & (ULONG) ~(1 << 31));
		}
	}

	return resident;
}


/* Case-insensitive FindName()
 */
static
struct Node *findname(struct List *list, CONST_STRPTR name)
{
	struct Node *node;

	ForeachNode(list, node)
	{
		if (!Stricmp(node->ln_Name, (STRPTR) name))
		{
			return node;
		}
	}

	return NULL;
}


/* Retrieve information from resident modules. Returns 0 for success.
 */
static
int createresidentver(struct Resident *MyResident)
{
	STRPTR buffer = NULL, tmpbuffer;
	int error, pos = 0, foundver = FALSE;

	if (MyResident->rt_IdString)
	{
		buffer = skipwhites(MyResident->rt_IdString);
		//Printf("createresidentver: buffer \"%s\"\n", (LONG) buffer);

		/* Locate version part */
		while (buffer[pos])
		{
			LONG dummy;

			/* NOTE: Not isspace()! - Piru */
			if (myisspace(buffer[pos]) &&
			    StrToLong(buffer + pos + 1, &dummy) != -1)
			{
				buffer += pos;
				foundver = TRUE;
				break;
			}
			pos++;
		}
		//Printf("createresidentver: buffer: \"%s\"\n", (LONG) buffer);
	}


	/* If could not find any version info, use the resident rt_Name */
	if (!foundver)
		buffer = "";

	//Printf("createresidentver: buffer: \"%s\"\n", (LONG) buffer);

	tmpbuffer = AllocVec(strlen(MyResident->rt_Name) + strlen(buffer) + 1, MEMF_ANY);
	if (!tmpbuffer)
	{
		PrintFault(ERROR_NO_FREE_STORE, (STRPTR) ERROR_HEADER);
		return RETURN_FAIL;
	}

	strcpy(tmpbuffer, MyResident->rt_Name);
	strcat(tmpbuffer, buffer);
	//Printf("createresidentver: tmpbuffer: \"%s\"\n", (LONG) tmpbuffer);
	error = makedatafromstring(tmpbuffer);

	FreeVec(tmpbuffer);

	return error;
}


/* Retrieve version information from library. Returns 0 for success.
 */
static
int createlibraryver(struct Library *MyLibrary)
{
	STRPTR buffer, tmpbuffer;
	int error, foundver = FALSE, pos;

	if (MyLibrary->lib_IdString)
	{
		//Printf("createlibraryver: lib_IdString \"%s\"\n", (LONG) MyLibrary->lib_IdString);
		buffer = skipwhites(MyLibrary->lib_IdString);

		//Printf("createlibraryver: buffer \"%s\"\n", (LONG) buffer);

		/* Find full 'ver.rev' version info
		 */
		pos = 0;
		while (buffer[pos])
		{
			LONG dummy, add;

			/* NOTE: Not isspace()! - Piru */
			if (myisspace(buffer[pos]) &&
			    (add = StrToLong(buffer + pos + 1, &dummy)) != -1 &&
				buffer[pos + 1 + add] == '.')
			{
				buffer += pos;
				pos = 0;
				foundver = TRUE;
				break;
			}
			pos++;
		}

		/* If could not find 'ver.rev', find any numeric */
		if (!foundver)
		{
			pos = 0;
			while (buffer[pos])
			{
				LONG dummy;

				/* NOTE: Not isspace()! - Piru */
				if (myisspace(buffer[pos]) &&
				    StrToLong(buffer + pos + 1, &dummy) != -1)
				{
					buffer += pos;
					pos = 0;
					foundver = TRUE;
					break;
				}
				pos++;
			}
		}
	}

	/* If could not find any version info, use the resident rt_Name */
	if (!foundver)
	{
		buffer = "";
		error = RETURN_WARN;
	}

	tmpbuffer = AllocVec(strlen(MyLibrary->lib_Node.ln_Name) + strlen(buffer) + 1, MEMF_ANY);
	if (!tmpbuffer)
	{
		PrintFault(ERROR_NO_FREE_STORE, (STRPTR) ERROR_HEADER);
		return RETURN_FAIL;
	}

	strcpy(tmpbuffer, MyLibrary->lib_Node.ln_Name);
	strcat(tmpbuffer, buffer);
	//Printf("createlibraryver: tmpbuffer: \"%s\"\n", (LONG) tmpbuffer);
	pos = makedatafromstring(tmpbuffer);
	if (pos > error)
	{
		error = pos;
	}

	FreeVec(tmpbuffer);

	return error;
}


/* Create default strings
 */
static
int createdefvers(CONST_STRPTR name)
{
	FreeVec(parsedver.pv_revname);
	FreeVec(parsedver.pv_vername);
	FreeVec(parsedver.pv_name);

	parsedver.pv_name    = dupstr(name, -1);
	parsedver.pv_vername = AllocVec(14, MEMF_ANY);
	parsedver.pv_revname = AllocVec(15, MEMF_ANY);

	if (parsedver.pv_name &&
	    parsedver.pv_vername &&
	    parsedver.pv_revname)
	{
		sprintf(parsedver.pv_vername, "%ld", parsedver.pv_version);
		sprintf(parsedver.pv_revname, ".%ld", parsedver.pv_revision);

		return RETURN_OK;
	}

	return RETURN_FAIL;
}


/* Create version info from named resident
 */
static
int makeresidentver(CONST_STRPTR name)
{
	struct Resident *MyResident;
	int error = -1;

	if ((MyResident = findresident(name)))
	{
		error = createresidentver(MyResident);
		if (error != RETURN_OK)
		{
			/* get values from residenttag */
			parsedver.pv_version  = MyResident->rt_Version;
			parsedver.pv_revision = -1;

#ifdef __MORPHOS__
			if (MyResident->rt_Flags & RTF_EXTENDED)
			{
				parsedver.pv_revision = MyResident->rt_Revision;
			}
#endif /* __MORPHOS__ */

			error = createdefvers(MyResident->rt_Name);
		}
	}

	return error;
}


/* Create version info from named list node
*/
static
int makeexeclistver(struct List *list, CONST_STRPTR name)
{
	struct Library *MyLibrary;
	int error = -1;

	Forbid();

	MyLibrary = (struct Library *) findname(list, name);
	if (MyLibrary)
	{
		/* get values from library */
		ULONG ver = MyLibrary->lib_Version;
		ULONG rev = MyLibrary->lib_Revision;

		error = createlibraryver(MyLibrary);
		if (error != RETURN_OK ||
		    parsedver.pv_version != ver ||
		    parsedver.pv_revision != rev)
		{
			/* special case where createlibraryrev was successful, but
			 * version or revision don't match.
			 */
			if (error == RETURN_OK)
			{
				/* If there is extrastr, make sure there's linefeed, too.
				 */
				if (parsedver.pv_extrastr)
				{
					parsedver.pv_extralf = "\n";
				}
			}

			/* get values from library */
			parsedver.pv_version  = ver;
			parsedver.pv_revision = rev;

			error = createdefvers(MyLibrary->lib_Node.ln_Name);
		}
	}

	Permit();

	return error;
}


/* Find resident from seglist, return NULL if none found
*/
static
struct Resident *FindLibResident(BPTR	Segment)
{
	while (Segment)
	{
		ULONG		*MySegment;
		UWORD		*MyBuffer;
		UWORD		*EndBuffer;

		MySegment	= (ULONG*) BADDR(Segment);
		MyBuffer	= (UWORD*) &MySegment[1];
		EndBuffer	= (UWORD*) &MySegment[(MySegment[-1] - sizeof(ULONG) * 4) / sizeof(ULONG)];

		while (MyBuffer < EndBuffer)
		{
			struct Resident	*MyResident;

			MyResident = (struct Resident*) MyBuffer;
			if (MyResident->rt_MatchWord == RTC_MATCHWORD &&
			    MyResident->rt_MatchTag == MyResident)
			{
				return MyResident;
			}

			MyBuffer++;
		}

		Segment	=(BPTR) MySegment[0];
	}

	SetIoErr(ERROR_OBJECT_NOT_FOUND);
	return NULL;
}


/* Find $VER: tag from seglist, return NULL if none found
   Returns dupstr()d string or NULL.
*/
static
STRPTR FindSegmentVER(BPTR	Segment)
{
	while (Segment)
	{
		ULONG		*MySegment;
		CONST_STRPTR	MyBuffer;
		CONST_STRPTR	EndBuffer;
		CONST_STRPTR	SegmentEnd;

		MySegment	= (ULONG*) BADDR(Segment);
		MyBuffer	= (CONST_STRPTR) &MySegment[1];
		SegmentEnd	= ((CONST_STRPTR) MySegment) + MySegment[-1] - sizeof(ULONG) * 2;
		EndBuffer	= SegmentEnd - 5;

		while (MyBuffer < EndBuffer)
		{
			if (MyBuffer[0] == '$' &&
			    MyBuffer[1] == 'V' &&
			    MyBuffer[2] == 'E' &&
			    MyBuffer[3] == 'R' &&
			    MyBuffer[4] == ':')
			{
				CONST_STRPTR EndPtr;

				MyBuffer += 5;
				/* Required because some smartass could end his $VER: tag
				 * without '\0' in the segment to save space. - Piru
				 */
				for (EndPtr = MyBuffer; EndPtr < SegmentEnd && *EndPtr; EndPtr++)
					;
				if (EndPtr - MyBuffer)
				{
					return dupstr(MyBuffer, EndPtr - MyBuffer);
				}
			}

			MyBuffer++;
		}

		Segment	=(BPTR) MySegment[0];
	}

	SetIoErr(ERROR_OBJECT_NOT_FOUND);
	return NULL;
}


/* Create version info from named device
*/
static
int makedevicever(CONST_STRPTR name)
{
	struct DevProc *MyDevProc;
	int error = -1;

	MyDevProc = GetDeviceProc((STRPTR) name, NULL);
	if (MyDevProc)
	{
		struct DosList *devnode = NULL;
		BOOL doslistlocked = FALSE;

		if (MyDevProc->dvp_DevNode->dol_Type == DLT_DEVICE)
		{
			devnode = MyDevProc->dvp_DevNode;
		}
		else if (MyDevProc->dvp_DevNode->dol_Type == DLT_VOLUME)
		{
			struct DosList *dl;

			dl = LockDosList(LDF_DEVICES | LDF_READ);
			while ((dl = NextDosEntry(dl, LDF_DEVICES)))
			{
				if (dl->dol_Task == MyDevProc->dvp_DevNode->dol_Task)
				{
					devnode = dl;
					break;
				}
			}
			if (devnode)
			{
				doslistlocked = TRUE;
			}
			else
			{
				UnLockDosList(LDF_DEVICES | LDF_READ);
			}
		}

		if (devnode)
		{
			BPTR SegList;

			SegList = devnode->dol_misc.dol_handler.dol_SegList;
			if (SegList)
			{
				struct Resident *MyResident;

				MyResident = FindLibResident(SegList);
				if (MyResident)
				{
					error = createresidentver(MyResident);
				}
				else
					error = RETURN_FAIL;
			}
		}

		if (doslistlocked)
		{
			UnLockDosList(LDF_DEVICES | LDF_READ);
		}

		FreeDeviceProc(MyDevProc);
	}

	if (error != RETURN_OK && error != -1)
	{
		Printf("Could not find version information for '%s'\n", (LONG) name);
	}

	return error;
}


/* Retrieve version information from file. Return 0 for success.
 */
#define BUFFERSIZE (16384 + 1)
static
int makefilever(CONST_STRPTR name)
{
	BPTR file;
	int error; // = RETURN_OK;

	file = Open((STRPTR) name, MODE_OLDFILE);
	if (file)
	{
		UBYTE *buffer;

		buffer = AllocMem(BUFFERSIZE, MEMF_PUBLIC);
		if (buffer)
		{
			int len = BUFFERSIZE - 1;

			error = findinfile(file, "$VER:", buffer, &len, parsedver.pv_md5sum);
			if (error == RETURN_OK)
			{
				parsedver.pv_flags |= PVF_MD5SUM;

				if (len >= 0)
				{
					STRPTR startbuffer;

					buffer[len] = '\0';
					startbuffer = skipwhites(buffer);

					//Printf("startbuffer \"%s\"\n", startbuffer);
					error = makedatafromstring(startbuffer);
				}
				else
				{
					error = RETURN_ERROR;

					if (IsFileSystem((STRPTR) name))
					{
						/* Try LoadSeg
						 */
						Close(file);

						file = LoadSeg((STRPTR) name);
						if (file)
						{
							struct Resident *MyResident;

							MyResident = FindLibResident(file);
							if (MyResident /*&&
							    (MyResident->rt_Type == NT_LIBRARY ||
							     MyResident->rt_Type == NT_DEVICE)*/)
							{
								error = createresidentver(MyResident);
							}

							UnLoadSeg(file);
						}
						file = NULL;
					}

					if (error != RETURN_OK)
					{
						/* If user didn't ask for md5sum or we could not calculate it.
						*/
						if (!args.arg_md5sum || (!(parsedver.pv_flags & PVF_MD5SUM)))
						{
							Printf("Could not find version information for '%s'\n", (LONG) name);
						}
					}
				}
			}
			else
			{
				PrintFault(IoErr(), (STRPTR) ERROR_HEADER);
			}

			FreeMem(buffer, BUFFERSIZE);
		}
		else
		{
			error = RETURN_FAIL;
			PrintFault(IoErr(), (STRPTR) ERROR_HEADER);
		}

		if (file)
			Close(file);
	}
	else
	{
		LONG ioerr = IoErr();

		if (ioerr == ERROR_OBJECT_NOT_FOUND ||
		    ioerr == ERROR_OBJECT_WRONG_TYPE)
		{
			error = -1;
		}
		else
		{
			PrintFault(IoErr(), (STRPTR) ERROR_HEADER);
			error = RETURN_FAIL;
		}
	}

	return error;
}


static
int makerescmdver(CONST_STRPTR name)
{
	int error = -1;
	struct Segment *segment;

	Forbid();

	segment = FindSegment((STRPTR) name, NULL, 0);
	if (!segment)
	{
		segment = FindSegment((STRPTR) name, NULL, 1);
	}

	if (segment)
	{
		if (segment->seg_UC == CMD_INTERNAL ||
		    segment->seg_UC == CMD_DISABLED)
		{
			Permit();

#ifdef __MORPHOS__

			/* Another reason why I separated resident commands from
			 * shell... :) - Piru
			 */
			error = makeresidentver("shellcmd");

#else /* __MORPHOS__ */

			error = makeresidentver("shell");

#endif /* __MORPHOS__ */

			Forbid();
		}
		else
		{
			STRPTR buffer = FindSegmentVER(segment->seg_Seg);
			if (buffer)
			{
				STRPTR startbuffer;

				startbuffer = skipwhites(buffer);

				//Printf("startbuffer \"%s\"\n", (LONG) startbuffer);
				error = makedatafromstring(startbuffer);

				FreeVec(buffer);
			}
		}
	}

	Permit();

	return error;
}

static
int setvervar(CONST_STRPTR name, LONG ver, LONG rev)
{
	UBYTE buf[32];

	sprintf(buf, "%ld.%ld", (LONG) ver, (LONG) rev);

	return SetVar((STRPTR) name, buf, -1, GVF_LOCAL_ONLY | LV_VAR) ? RETURN_OK : -1;
}


#ifdef __MORPHOS__

static
void doambientverstring(void)
{
	setvervar("Ambient",
	          parsedver.pv_version,
	          parsedver.pv_revision);

	Printf(", Ambient %ld.%ld, ", parsedver.pv_version, parsedver.pv_revision);
}

static
int doambientver(CONST_STRPTR name)
{
	int error;

	error = makefilever(name);
	if (error == RETURN_OK)
	{
		doambientverstring();
	}

	freeverstring();

	return error;
}

static
STRPTR DoARexxCmd(CONST_STRPTR port, CONST_STRPTR command)
{
  STRPTR res = NULL;
  struct Library *RexxSysBase;

  if ((RexxSysBase = OpenLibrary("rexxsyslib.library", 36UL)))
  {
    struct MsgPort *ReplyPort;

    if ((ReplyPort = CreateMsgPort()))
    {
      struct RexxMsg *HostMsg;

      if ((HostMsg = CreateRexxMsg(ReplyPort, NULL, port)))
      {
        if ((HostMsg->rm_Args[0] = CreateArgstring(command, strlen(command))))
        {
          struct MsgPort *RexxPort;

          HostMsg->rm_Action = RXCOMM | RXFF_RESULT;

          Forbid();
          if ((RexxPort = FindPort(port)))
          {
            struct RexxMsg *answer;

            PutMsg(RexxPort, (struct Message *)HostMsg);
            Permit();

            do
            {
              WaitPort(ReplyPort);
              answer = (struct RexxMsg *)GetMsg(ReplyPort);
            } while (!answer);
            
            if (answer->rm_Result1 == 0)
            {
              if (answer->rm_Result2)
              {
                ULONG size = strlen((char *) answer->rm_Result2) + 1;
                res = AllocVec(size, MEMF_ANY);
                if (res)
                {
                  memcpy(res, (void *) answer->rm_Result2, size);
                }
                DeleteArgstring((UBYTE *)answer->rm_Result2);
              }
            }
          }
          else
          {
            Permit();
          }              

          DeleteArgstring(HostMsg->rm_Args[0]);
        }

        DeleteRexxMsg(HostMsg);
      }

      DeleteMsgPort(ReplyPort);
    }

    CloseLibrary(RexxSysBase);
  }

  return res;
}

static
int makeambientversion(void)
{
	int error = -1;
	UBYTE path[256];
	struct Process *Self = (struct Process *) FindTask(NULL);
	APTR oldwinptr;
	STRPTR str;

	oldwinptr = Self->pr_WindowPtr;
	Self->pr_WindowPtr = (APTR) -1;

	str = DoARexxCmd("AMBIENT", "VERSION");
	if (str)
	{
		LONG num;

		num = StrToLong(str, &parsedver.pv_version);
		if (num > 0 && str[num] == '.')
		{
			if (StrToLong(str + num + 1, &parsedver.pv_revision) > 0)
			{
				doambientverstring();
				parsedver.pv_version = 0;
				parsedver.pv_revision = 0;
				error = 0;
			}
		}

		FreeVec(str);
	}

	/* If no version could be queried via AMBIENT ARexx port,
	   scan the binary directly. */
	if (error)
	{
		if (GetVar("ambient_path", path, sizeof(path), LV_VAR) >= 0)
		{
			error = doambientver(path);
		}
		if (error)
		{
			error = doambientver("mossys:ambient/ambient");
			if (error)
			{
				error = doambientver("sys:system/ambient/ambient");
			}
		}
	}

	Self->pr_WindowPtr = oldwinptr;

	return error;
}

#endif /* __MORPHOS__ */


static
int makekickversion(void)
{
	parsedver.pv_version  = SysBase->LibNode.lib_Version;
	parsedver.pv_revision = SysBase->SoftVer;

	setvervar("Kickstart",
	          parsedver.pv_version,
	          parsedver.pv_revision);

	Printf("Kickstart %ld.%ld",
	       (LONG) parsedver.pv_version, (LONG) parsedver.pv_revision);

	return RETURN_OK;
}


static
int makewbversion(void)
{
	int error = -1;
	struct Library *VersionBase;

	VersionBase = OpenLibrary("version.library", 0);
	if (VersionBase)
	{
		error = makeexeclistver(&SysBase->LibList, "version.library");

		if (error == RETURN_OK)
		{
			STRPTR newname = dupstr("Workbench", -1);
			if (newname)
			{
				FreeVec(parsedver.pv_name);
				parsedver.pv_name = newname;
			}
			setvervar("Workbench", parsedver.pv_version, parsedver.pv_revision);
		}

		CloseLibrary(VersionBase);
	}

	return error;
}


static
int makesysver(void)
{
	int error;

#ifdef __MORPHOS__
	int ambient = -1;

	error = makeresidentver("MorphOS");
	if (error == RETURN_OK)
	{
		Printf("MorphOS %ld.%ld",
		       parsedver.pv_version,
		       parsedver.pv_revision);

		setvervar("MorphOS",
		          parsedver.pv_version,
		          parsedver.pv_revision);

		ambient = makeambientversion();
		if (ambient != RETURN_OK)
			PutStr(", ");
	}

#endif /* __MORPHOS__ */

	error = makekickversion();
	if (error == RETURN_OK)
	{
		error = makewbversion();
#ifdef __MORPHOS__
		/* Don't output the Workbench version if we have ambient version already */
		if (ambient != -1)
		{
			PutStr("\n");
			error = RETURN_WARN;
		}
#endif
		if (error == RETURN_OK)
		{
			PutStr(", ");
		}
	}

	return error;
}


/* Determine, by which means to get the version-string.
 */
static
int makeverstring(CONST_STRPTR name)
{
	int error; // = RETURN_OK;
	BOOL volume = name[strlen(name) - 1] == ':';
	CONST_STRPTR filepart = FilePart(name);

	error = -1;

	if (!volume && !args.arg_file)
	{
		if (*filepart)
		{
			error = makeresidentver(filepart);
			if (error != RETURN_OK)
			{

				/* Try libraries
				 */
				error = makeexeclistver(&SysBase->LibList, filepart);
				if (error != RETURN_OK)
				{
					STRPTR namebuf;
					ULONG namelen = strlen(filepart);
#ifdef __MORPHOS__
					ULONG elf = ((namelen > 4) ? ((*((ULONG *)(filepart + namelen - 4)) == 0x2E656C66) ? 4 : 0) : 0);	/* Long live big-endian hacks! ;) */
#endif

					/* 12 is "MOSSYS:LIBS/" */
					if ((namebuf = AllocVec(12 + namelen + 4 + 1, MEMF_PUBLIC)))
					{
#ifdef __MORPHOS__
						strcpy(namebuf, "MOSSYS:LIBS/");
						strcat(namebuf, filepart);
						if (!elf) strcat(namebuf, ".elf");
						error = makefilever(namebuf);

						if (error != RETURN_OK)
						{
							namebuf[12 + namelen - elf] = '\0';
							error = makefilever(namebuf);

							if (error != RETURN_OK)
							{
								strcpy(namebuf, "LIBS:");
								strcat(namebuf, filepart);
								if (!elf) strcat(namebuf, ".elf");
								error = makefilever(namebuf);

								if (error != RETURN_OK)
								{
									namebuf[5 + namelen - elf] = '\0';
									error = makefilever(namebuf);

									if (error != RETURN_OK)
									{
										/* Try devices
										 */
										error = makeexeclistver(&SysBase->DeviceList, filepart);
										if (error != RETURN_OK)
										{
											strcpy(namebuf, "MOSSYS:DEVS/");
											strcat(namebuf, filepart);
											if (!elf) strcat(namebuf, ".elf");
											error = makefilever(namebuf);

											if (error != RETURN_OK)
											{
												namebuf[12 + namelen - elf] = '\0';
												error = makefilever(namebuf);

												if (error != RETURN_OK)
												{
													strcpy(namebuf, "DEVS:");
													strcat(namebuf, filepart);
													if (!elf) strcat(namebuf, ".elf");
													error = makefilever(namebuf);

													if (error != RETURN_OK)
													{
														namebuf[5 + namelen - elf] = '\0';
														error = makefilever(namebuf);
													}
												}
											}
										}
									}
								}
							}
						}

#else

						strcpy(namebuf, "LIBS:");
						strcat(namebuf, filepart);
						error = makefilever(namebuf);

						/* Try devices
						*/
						if (error != RETURN_OK)
						{
							error = makeexeclistver(&SysBase->DeviceList, filepart);
							if (error != RETURN_OK)
							{
								strcpy(namebuf, "DEVS:");
								strcat(namebuf, filepart);
								error = makefilever(namebuf);
							}
						}

#endif

						FreeVec(namebuf);
					}
				}
			}
		}
	}

	if (!args.arg_res && error == -1)
	{
		if (volume)
		{
			error = makedevicever(name);
		}
		else
		{
			if (*filepart)
			{
				error = makefilever(name);
			}
		}
	}

	if (!args.arg_file && error == -1)
	{
		error = makerescmdver(name);
	}

	if (error)
	{
		/* If user asked for md5sum, and we could calculate it, don't print error
		 *  but the md5sum + file.
		 */
		if (args.arg_md5sum && (parsedver.pv_flags & PVF_MD5SUM))
		{
			parsedver.pv_name = dupstr(name, -1);
			parsedver.pv_flags |= PVF_NOVERSION;
			error = RETURN_OK;
		}
	}

	if (error == -1)
	{
		PrintFault(ERROR_OBJECT_NOT_FOUND, (STRPTR) ERROR_HEADER);
		error = RETURN_FAIL;
	}

	return error;
}


static
void freeverstring(void)
{
	parsedver.pv_flags    = 0;
	parsedver.pv_version  = 0;
	parsedver.pv_revision = 0;

	FreeVec(parsedver.pv_extrastr);
	parsedver.pv_extrastr = NULL;
	parsedver.pv_extralf = NULL;
	FreeVec(parsedver.pv_datestr);
	parsedver.pv_datestr  = NULL;
	FreeVec(parsedver.pv_revname);
	parsedver.pv_revname  = NULL;
	FreeVec(parsedver.pv_vername);
	parsedver.pv_vername  = NULL;
	FreeVec(parsedver.pv_name);
	parsedver.pv_name     = NULL;
}

/* Compare the version given as argument with the version from the object.
 * Return RETURN_WARN, if args-v>object-v, otherwise return RETURN_OK.
 */
static
int cmpargsparsed(void)
{
	if (args.arg_version)
	{
		if (*(args.arg_version) > parsedver.pv_version)
		{
			return RETURN_WARN;
		}
		else if (*(args.arg_version) == parsedver.pv_version && args.arg_revision)
		{
			if (*(args.arg_revision) > parsedver.pv_revision)
			{
				return RETURN_WARN;
			}
		}
	}
	else if (args.arg_revision)
	{
		if (*(args.arg_revision) > parsedver.pv_revision)
		{
			return RETURN_WARN;
		}
	}
	return RETURN_OK;
}
